#ifndef ITC_H
#define ITC_H

#include <iostream>
#include <string>
#include <queue>
#include <cassert>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

namespace itc {

typedef boost::function<void (void)> CallType;

class WaitAndExecute
{
public:
	WaitAndExecute():
		m_mutex(), m_cond(), m_req(false), m_callQ()
	{}	

	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
		m_req = true;
		m_cond.notify_one();
	}
	
	void execute()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		
		// wait
		while (!m_req)
		{
			m_cond.wait(lock);
		}
		
		// execute
		while (!m_callQ.empty())
		{
			m_callQ.front()();
			m_callQ.pop();
		}
		m_req = false;	
	}	

private:
	boost::mutex m_mutex;
	boost::condition_variable m_cond;
	bool m_req;
	std::queue<CallType>  m_callQ;	
};

class ProcessAndExecute
{
public:
	ProcessAndExecute() : 
		m_mutex(), m_callQ()
	{}
	
	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
	}
	
	void execute()
	{
		CallType call;
		bool req = false;
		
		// process
		process();
		
		// execute
		{
			boost::unique_lock<boost::mutex> lock(m_mutex);
			if (!m_callQ.empty())
			{
				call = m_callQ.front();
				m_callQ.pop();
				req = true;
			}
		}
		
		if (req)
			call();		
	}	

private:
	virtual void process() = 0;

	boost::mutex m_mutex;
	std::queue<CallType>  m_callQ;
};

template <typename Callee, typename PolicyType = WaitAndExecute>
class ITCThread
{
public:
	ITCThread(Callee* callee, const std::string name = "itcthr") :
		m_callee(callee), m_thr(), m_policy(new PolicyType),
		m_name(name), m_run(true), m_code(0)
	{
		m_thr = boost::thread(&ITCThread::run, this);
	}

	const std::string& getName()
	{
		return m_name;
	}

	void quit(int code = 0)
	{
		doItPrivate(&ITCThread::doQuit, code);
	}

	int join()
	{
		m_thr.join();
		return m_code;
	}

	void doIt(CallType call)
	{
		m_policy->queue(call);
	}

	void doIt(void (Callee::*mf)(void))
	{
		assert(m_callee != 0);
		doIt(CallType(boost::bind(mf, m_callee)));
	}

	void doIt(void (*func)(void))
	{
		doIt(CallType(func));
	}

	template <typename ArgType>
	void doIt(void (Callee::*mf)(ArgType), ArgType arg)
	{
		assert(m_callee != 0);
		doIt(CallType(boost::bind(mf, m_callee, arg)));
	}

	template <typename ArgType>
	void doIt(void (*func)(ArgType), ArgType arg)
	{
		doIt(CallType(boost::bind(func, arg)));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
	{
		assert(m_callee != 0);
		doIt(CallType(boost::bind(mf, m_callee, arg1, arg2)));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (*func)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
	{
		doIt(CallType(boost::bind(func, arg1, arg2)));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		assert(m_callee != 0);
		doIt(CallType(boost::bind(mf, m_callee, arg1, arg2, arg3)));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (*func)(ArgType1, ArgType2, ArgType3), 
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		doIt(CallType(boost::bind(func, arg1, arg2, arg3)));
	}

private:
	void doItPrivate(void (ITCThread::*mf)(int), int arg)
	{
		doIt(CallType(boost::bind(mf, this, arg)));
	}

	void run()
	{
		std::cout << m_name << ": starting..." << std::endl;
		while (m_run)
		{
			m_policy->execute();
		}
	}

	void doQuit(int code)
	{
		m_run = false;
		m_code = code;
	}

	Callee* m_callee;
	boost::thread m_thr;
	boost::shared_ptr<PolicyType> m_policy;
	const std::string m_name;
	bool m_run;
	int m_code;
};

struct Nothing {};

template <typename PolicyType = WaitAndExecute>
struct FunctionITCAdaptor : public ITCThread<Nothing, PolicyType>
{
	FunctionITCAdaptor(const std::string& name) :
		ITCThread<Nothing, PolicyType>(0, name) {}
};

};

#endif
